// Project Sword & Gun Copyright © 2021 Kevin Poretti


#include "Components/SNGCharacterMovementComponent.h"

#include "GameFramework/Character.h"
#include "Kismet/KismetMathLibrary.h"
#include "PhysicsEngine/PhysicsSettings.h"

USNGCharacterMovementComponent::USNGCharacterMovementComponent()
{
	// Defaults
	DefaultGroundFriction = 8.0f;
	DefaultMaxWalkSpeed = 600.0f;
	DefaultRotationRate = FRotator(0.0f, 360.0f, 0.0f);
	
	// sprint
	SprintingMaxWalkSpeed = 1000.0f;
	SprintLateralSpeedThreshold = 500.0f;
	SprintingRotationRate = FRotator(0.0f, 180.0f, 0.0f);
	TimeToInitiateSprint = 2.0f;
	WalkToSprintInterpSpeed = 4.0f;

	// firing
	FiringMaxWalkSpeed = 300.0f;
	
	// reloading
	ReloadingMaxWalkSpeed = 300.0f;
	
	// airborne defaults
	DefaultAirControl = 0.3f;
	
	// jump
	MaxJumpHeight = 250.0f;
	DesiredJumpDuration = 0.75f;
	ComputeGravityScaleAndJumpZVelocity();

	// dash
	DashVelocity = 1000.0f;
	DashGroundFriction = 0.0f;
	DashGravityScale = 0.1f;
	DashFloatTime = 0.2f;
	DashSlideTime = 0.5f;
	DashGravityToDefaultSpeed = 5.0f;
	DashGroundFrictionToDefaultSpeed = 5.0f;
	
	/*
	// not used...yet
	MinJumpHeight = 200.0f;
	MaxJumpTime = 0.2f;

	LateralVelocityDampening = 0.8f;
	
	MinAllowedDashHeight = 175.0f;
	MinAllowedAttackHeight = 175.0f;
	*/

	// Character Movement component defaults
	SetWalkableFloorAngle(45.622231f);
	
	AirControlBoostMultiplier = 0.0f;
	AirControlBoostVelocityThreshold = 0.0f;
}

/**
 *	Instantly sets all movement parameters to their defaults values (i.e. ground friction to DefaultGroundFriction,
 *	MaxWalkSpeed to DefaultMaxWalkSpeed, etc.)
 */
void USNGCharacterMovementComponent::RestoreMovementDefaults()
{
	GroundFriction = TargetGroundFriction = DefaultGroundFriction;
	GravityScale = TargetGravityScale = DefaultGravityScale;
	RotationRate = DefaultRotationRate;
	MaxWalkSpeed = TargetMaxWalkSpeed = DefaultMaxWalkSpeed;
	AirControl = DefaultAirControl;
	JumpZVelocity = DefaultJumpZVelocity;
}

void USNGCharacterMovementComponent::Dash()
{
	GroundFriction = DashGroundFriction;
	GravityScale = DashGravityScale;
	FloatTime = DashFloatTime;
	SlideTime = DashSlideTime;

	// sets gravity and friction targets which will start interpolating after float time and slide time elapse respectively
	TargetGravityScale = DefaultGravityScale;
	TargetGroundFriction = DefaultGroundFriction;
	GravityScaleInterpSpeed = DashGravityToDefaultSpeed;
	GroundFrictionInterpSpeed = DashGroundFrictionToDefaultSpeed;
	
	FRotator CharacterRot = FRotator(0.0f, GetCharacterOwner()->GetActorRotation().Yaw, 0.0f);
	FVector LaunchDir = FVector(UKismetMathLibrary::GetForwardVector(CharacterRot));
	LaunchDir.Normalize();
	Launch(LaunchDir * DashVelocity);
}

#if WITH_EDITOR
void USNGCharacterMovementComponent::PostEditChangeProperty(FPropertyChangedEvent& PropertyChangedEvent)
{
	Super::PostEditChangeProperty(PropertyChangedEvent);

	const FProperty* PropertyThatChanged = PropertyChangedEvent.MemberProperty;
	if (PropertyThatChanged && PropertyThatChanged->GetFName() == GET_MEMBER_NAME_CHECKED(USNGCharacterMovementComponent, MaxJumpHeight) ||
		PropertyThatChanged && PropertyThatChanged->GetFName() == GET_MEMBER_NAME_CHECKED(USNGCharacterMovementComponent, DesiredJumpDuration))
	{
		// Compute WalkableFloorZ from the Angle.
		ComputeGravityScaleAndJumpZVelocity();
	}
}

void USNGCharacterMovementComponent::PostLoad()
{
	Super::PostLoad();

	// Compute gravity scale and jump z velocity for first time
	ComputeGravityScaleAndJumpZVelocity();
}
#endif // WITH_EDITOR

void USNGCharacterMovementComponent::OnSprintTimer()
{
	bIsSprinting = true;
	TargetMaxWalkSpeed = SprintingMaxWalkSpeed;
}

void USNGCharacterMovementComponent::ComputeGravityScaleAndJumpZVelocity()
{
	float HalfJumpDuration = DesiredJumpDuration / 2.0f;
	DefaultJumpZVelocity = (2 * MaxJumpHeight) / HalfJumpDuration;
	
	float GameGravity = UPhysicsSettings::Get()->DefaultGravityZ;
	float TargetGravity = (-2 * MaxJumpHeight) / (HalfJumpDuration * HalfJumpDuration);
	DefaultGravityScale = TargetGravity / GameGravity;
}

void USNGCharacterMovementComponent::StartJump()
{
	/*
	JumpTimer = MaxJumpTime;

	Velocity = FVector(Velocity.X * LateralVelocityDampening, Velocity.Y * LateralVelocityDampening, 0.0f);
	*/
}

void USNGCharacterMovementComponent::AddJumpForce()
{
	/*
	if(JumpTimer > 0.0f)
	{
		Velocity = FVector(Velocity.X, Velocity.Y, Velocity.Z + (JumpForce * GetWorld()->DeltaTimeSeconds));

		JumpTimer = FMath::Max(JumpTimer - GetWorld()->DeltaTimeSeconds, 0.0f);
	}
	*/
}

void USNGCharacterMovementComponent::InterpolateMovementParameters(float DeltaTime)
{
	// set walk speed
	MaxWalkSpeed = FMath::FInterpTo(MaxWalkSpeed, TargetMaxWalkSpeed, DeltaTime, MaxWalkSpeedInterpSpeed);

	// ground friction
	if(SlideTime > 0.0f)
	{
		SlideTime = FMath::Max(SlideTime - DeltaTime * GetCharacterOwner()->CustomTimeDilation, 0.0f);
	}
	else
	{
		GroundFriction = FMath::FInterpTo(GroundFriction, TargetGroundFriction, DeltaTime, GroundFrictionInterpSpeed);
	}

	// gravity scale
	if(FloatTime > 0.0f)
	{
		FloatTime = FMath::Max(FloatTime - DeltaTime * GetCharacterOwner()->CustomTimeDilation, 0.0f);
	}
	else
	{
		GravityScale = FMath::FInterpTo(GravityScale, TargetGravityScale, DeltaTime, GravityScaleInterpSpeed);
	}
	
	// set rotation rate
	// NOTE(kevin): since default rotate rate is 360 RInterpTo will go from 0 to 180 instead of 360 to 180
	// need to think of a way to prevent this
	// might just need to do my own interp manually
	/*
	FRotator TargetRotationRate = bIsSprinting ? SprintingRotationRate : DefaultRotationRate;
	FRotator NewRotationRate = FMath::RInterpTo(RotationRate, TargetRotationRate, DeltaTime, WalkToSprintInterpSpeed);
	RotationRate = NewRotationRate;
	*/

	// gravity
}

void USNGCharacterMovementComponent::BeginPlay()
{
	Super::BeginPlay();

	// set character movement parameters to defaults
	RestoreMovementDefaults();
}

void USNGCharacterMovementComponent::TickComponent(float DeltaTime, ELevelTick Tick,
	FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, Tick, ThisTickFunction);

	float LateralSpeed = Velocity.Size2D();
	if(LateralSpeed >= SprintLateralSpeedThreshold && !GetWorld()->GetTimerManager().IsTimerActive(TimerHandle_SprintTimer))
	{
		GetWorld()->GetTimerManager().SetTimer(TimerHandle_SprintTimer,
			this, &USNGCharacterMovementComponent::OnSprintTimer,
			TimeToInitiateSprint, false);
	}
	
	if(LateralSpeed < SprintLateralSpeedThreshold)
	{
		if(GetWorld()->GetTimerManager().IsTimerActive(TimerHandle_SprintTimer))
		{
			GetWorld()->GetTimerManager().ClearTimer(TimerHandle_SprintTimer);
		}
		bIsSprinting = false;
		TargetMaxWalkSpeed = DefaultMaxWalkSpeed;
		MaxWalkSpeedInterpSpeed = WalkToSprintInterpSpeed;
	}
	
	InterpolateMovementParameters(DeltaTime);
}
